@@ -41,84 +41,119 @@ except ImportError:
 try:
     import ssl
 except ImportError:
-    print("Python socket module was not compiled with SSL support. Aborting...")
-    sys.exit(1)
+    sys.exit("Python socket module was not compiled with SSL support. Aborting...")
 
 
-###############################################################################
+VERSION = "1.6.3"
 
-VERSION = "1.6.0"
 
-###############################################################################
+DEFAULT_RCFILE_PATH = "~/.ecmwfapirc"
+ANONYMOUS_APIKEY_VALUES = (
+    "anonymous",
+    "https://api.ecmwf.int/v1",
+    "anonymous@ecmwf.int",
+)
+
+
+class APIKeyNotFoundError(Exception):
+    pass
 
 
 class APIKeyFetchError(Exception):
     pass
 
 
-def _get_apikey_from_environ():
-    try:
-        key = os.environ["ECMWF_API_KEY"]
-        url = os.environ["ECMWF_API_URL"]
-        email = os.environ["ECMWF_API_EMAIL"]
-        return key, url, email
-    except KeyError:
-        raise APIKeyFetchError("ERROR: Could not get the API key from the environment")
+def get_apikey_values_from_environ():
+    apikey_values = (
+        os.getenv("ECMWF_API_KEY"),
+        os.getenv("ECMWF_API_URL"),
+        os.getenv("ECMWF_API_EMAIL"),
+    )
 
-
-def _get_apikey_from_rcfile():
-    if "ECMWF_API_RC_FILE" in os.environ:
-        rc = os.path.normpath(os.path.expanduser(os.environ["ECMWF_API_RC_FILE"]))
+    if not any(apikey_values):
+        raise APIKeyNotFoundError("ERROR: No API key found in the environment")
+    elif not all(apikey_values):
+        raise APIKeyFetchError("ERROR: Incomplete API key found in the environment")
     else:
-        rc = os.path.normpath(os.path.expanduser("~/.ecmwfapirc"))
+        return apikey_values
+
+
+def get_apikey_values_from_rcfile(rcfile_path):
+    rcfile_path = os.path.normpath(os.path.expanduser(rcfile_path))
 
     try:
-        with open(rc) as f:
-            config = json.load(f)
+        with open(rcfile_path) as f:
+            apikey = json.load(f)
+    except FileNotFoundError as e:
+        raise APIKeyNotFoundError(str(e))
     except IOError as e:  # Failed reading from file
         raise APIKeyFetchError(str(e))
     except ValueError:  # JSON decoding failed
-        raise APIKeyFetchError("ERROR: Missing or malformed API key in '%s'" % rc)
+        raise APIKeyFetchError(
+            "ERROR: Missing or malformed API key in '%s'" % rcfile_path
+        )
     except Exception as e:  # Unexpected error
         raise APIKeyFetchError(str(e))
-
-    try:
-        key = config["key"]
-        url = config["url"]
-        email = config["email"]
-        return key, url, email
-    except:
-        raise APIKeyFetchError("ERROR: Missing or malformed API key in '%s'" % rc)
+    else:
+        try:
+            return (apikey["key"], apikey["url"], apikey["email"])
+        except:
+            raise APIKeyFetchError(
+                "ERROR: Missing or malformed API key in '%s'" % rcfile_path
+            )
 
 
 def get_apikey_values():
-    """Get the API key from the environment or the '.ecmwfapirc' file.
+    """Get the API key values in Python tuple format either directly from the
+    environment, or from a file. If no API key is found, fall back to anonymous
+    access.
 
-    The environment is looked at first.
+    The complete workflow is the following:
+
+    - Step 1: the environment is checked for variables ECMWF_API_KEY,
+        ECMWF_API_URL, ECMWF_API_EMAIL.
+        * If all found, and not empty, return their values in Python tuple
+            format. 
+        * If only some found, and not empty, assume an incomplete API key, and
+            raise APIKeyFetchError.
+        * If none found, or found but empty, assume no API key available in the
+            environment, and continue to the next step.
+
+    - Step 2: the environment is checked for variable ECMWF_API_RC_FILE, meant
+        to point to a user defined API key file.
+        * If found, but pointing to a file not found, raise APIKeyNotFoundError.
+        * If found, and the file it points to exists, but cannot not be read, or
+            contains an invalid API key, raise APIKeyFetchError.
+        * If found, and the file it points to exists, can be read, and contains
+            a valid API key, return the API key in Python tuple format.
+        * If not found, or empty, assume no user provided API key file and
+            continue to the next step.
+
+    - Step 3: try the default ~/.ecmwfapirc file.
+        * Same as step 2, except for when ~/.ecmwfapirc is not found, where we
+            continue to the next step.
+
+    - Step 4: No API key found, so fall back to anonymous access.
 
     Returns:
-        Tuple with the key, url, and email forming our API key.
+        Pyhon tuple with the API key token, url, and email.
 
     Raises:
-        APIKeyFetchError: When unable to get the API key from either the
-            environment or the ecmwfapirc file.
+        APIKeyFetchError: If an API key is found, but invalid.
+        APIKeyNotFound: When ECMWF_API_RC_FILE is defined, but pointing to a
+            file that does not exist.
     """
     try:
-        key_values = _get_apikey_from_environ()
-    except APIKeyFetchError:
-        try:
-            key_values = _get_apikey_from_rcfile()
-        except APIKeyFetchError:
-            return (
-                os.environ.get("ECMWF_API_KEY", "anonymous"),
-                os.environ.get("ECMWF_API_URL", "https://api.ecmwf.int/v1"),
-                os.environ.get("ECMWF_API_EMAIL", "anonymous@ecmwf.int"),
-            )
-
-    return key_values
-
-
-###############################################################################
+        return get_apikey_values_from_environ()
+    except APIKeyNotFoundError:
+        env_rcfile_path = os.getenv("ECMWF_API_RC_FILE")
+        if env_rcfile_path:
+            return get_apikey_values_from_rcfile(env_rcfile_path)
+        else:
+            try:
+                return get_apikey_values_from_rcfile(DEFAULT_RCFILE_PATH)
+            except APIKeyNotFoundError:
+                return ANONYMOUS_APIKEY_VALUES
 
 
 class RetryError(Exception):
@@ -148,35 +183,35 @@ def robust(func):
                 return func(self, *args, **kwargs)
             except HTTPError as e:
                 if self.verbose:
-                    print("WARNING: HTTPError received %s" % (e))
+                    self.log("WARNING: HTTPError received %s" % e)
                 if e.code < 500 or e.code in (501,):  # 501: not implemented
                     raise
                 last_error = e
             except BadStatusLine as e:
                 if self.verbose:
-                    print("WARNING: BadStatusLine received %s" % (e))
+                    self.log("WARNING: BadStatusLine received %s" % e)
                 last_error = e
             except URLError as e:
                 if self.verbose:
-                    print("WARNING: URLError received %s %s" % (e.errno, e))
+                    self.log("WARNING: URLError received %s %s" % (e.errno, e))
                 last_error = e
             except APIException:
                 raise
             except RetryError as e:
                 if self.verbose:
-                    print("WARNING: HTTP received %s" % (e.code))
-                    print(e.text)
+                    self.log("WARNING: HTTP received %s" % e.code)
+                    self.log(e.text)
                 last_error = e
             except:
                 if self.verbose:
-                    print("Unexpected error:", sys.exc_info()[0])
-                    print(traceback.format_exc())
+                    self.log("Unexpected error: %s" % sys.exc_info()[0])
+                    self.log(traceback.format_exc())
                 raise
-            print("Error contacting the WebAPI, retrying in %d seconds ..." % delay)
+            self.log("Error contacting the WebAPI, retrying in %d seconds ..." % delay)
             time.sleep(delay)
             tries -= 1
         # if all retries have been exhausted, raise the last exception caught
-        print("Could not contact the WebAPI after %d tries, failing !" % max_tries)
+        self.log("Could not contact the WebAPI after %d tries, failing !" % max_tries)
         raise last_error
 
     return wrapped
@@ -189,26 +224,16 @@ def get_api_url(url):
     )
 
 
-SAY = True
-
-
 class Ignore303(HTTPRedirectHandler):
+
+    """Handler to automatically follow redirects.
+
+    Mainly implement when the API moved from http to https.
+    """
+
     def redirect_request(self, req, fp, code, msg, headers, newurl):
         if code in [301, 302]:
             # We want the posts to work even if we are redirected
-            if code == 301:
-                global SAY
-                if SAY:
-                    o = req.get_full_url()
-                    n = newurl
-                    print()
-                    print("*** ECMWF API has moved")
-                    print("***   OLD: %s" % get_api_url(o))
-                    print("***   NEW: %s" % get_api_url(n))
-                    print("*** Please update your ~/.ecmwfapirc file")
-                    print()
-                    SAY = False
-
             try:
                 # Python < 3.4
                 data = req.get_data()
@@ -233,14 +258,22 @@ class Ignore303(HTTPRedirectHandler):
         return None
 
     def http_error_303(self, req, fp, code, msg, headers):
-        infourl = addinfourl(fp, headers, req.get_full_url())
-        infourl.status = code
-        infourl.code = code
-        return infourl
+        return addinfourl(fp, headers, req.get_full_url(), code)
+
+
+def no_log(msg):
+    pass
+
+
+def print_with_timestamp(msg):
+    t = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
+    print("%s %s" % (t, msg))
 
 
 class Connection(object):
-    def __init__(self, url, email=None, key=None, verbose=False, quiet=False):
+    def __init__(
+        self, url, email=None, key=None, verbose=False, quiet=False, log=no_log
+    ):
         self.url = url
         self.email = email
         self.key = key
@@ -251,16 +284,16 @@ class Connection(object):
         self.offset = 0
         self.verbose = verbose
         self.quiet = quiet
+        self.log = log
         self.status = None
 
     @robust
     def call(self, url, payload=None, method="GET"):
-
         # Ensure full url
         url = urljoin(self.url, url)
 
         if self.verbose:
-            print(method, url)
+            self.log("Calling method %s on %s" % (method, url))
 
         headers = {
             "Accept": "application/json",
@@ -292,7 +325,7 @@ class Connection(object):
                     raise
         except HTTPError as e:
             if self.verbose:
-                print(e)
+                self.log(e)
             error = True
             res = e
             # 429: Too many requests
@@ -307,10 +340,10 @@ class Connection(object):
             self.location = urljoin(url, res.headers.get("Location", self.location))
 
         if self.verbose:
-            print("Code", code)
-            print("Content-Type", res.headers.get("Content-Type"))
-            print("Content-Length", res.headers.get("Content-Length"))
-            print("Location", res.headers.get("Location"))
+            self.log("Response code: %s" % code)
+            self.log("Response Content-Type: %s" % res.headers.get("Content-Type"))
+            self.log("Response Content-Length: %s" % res.headers.get("Content-Length"))
+            self.log("Response Location: %s" % res.headers.get("Location"))
 
         body = res.read().decode("utf-8")
         res.close()
@@ -326,17 +359,17 @@ class Connection(object):
                 error = True
 
         if self.verbose:
-            print(json.dumps(self.last, indent=4))
+            self.log("Response content: %s" % json.dumps(self.last, indent=4))
 
         self.status = self.last.get("status", self.status)
 
         if self.verbose:
-            print("Status", self.status)
+            self.log("Status %s" % self.status)
 
         if "messages" in self.last:
             for n in self.last["messages"]:
                 if not self.quiet:
-                    print(n)
+                    self.log(n)
                 self.offset += 1
 
         if code == 200 and self.status == "complete":
@@ -368,7 +401,7 @@ class Connection(object):
 
     def wait(self):
         if self.verbose:
-            print("Sleeping %s second(s)" % (self.retry))
+            self.log("Sleeping %s second(s)" % (self.retry))
         time.sleep(self.retry)
         self.call(self.location, None, "GET")
 
@@ -386,10 +419,6 @@ class Connection(object):
             pass
 
 
-def no_log(msg):
-    pass
-
-
 class APIRequest(object):
     def __init__(
         self,
@@ -404,21 +433,29 @@ class APIRequest(object):
     ):
         self.url = url
         self.service = service
-        self.connection = Connection(url, email, key, quiet=quiet, verbose=verbose)
-        self._log = log
+        self.log = log
         self.quiet = quiet
         self.verbose = verbose
-        self._empty_line = False
+
+        self.connection = Connection(
+            url, email=email, key=key, quiet=quiet, verbose=verbose, log=log
+        )
+
         self.log("ECMWF API python library %s" % (VERSION,))
         self.log("ECMWF API at %s" % (self.url,))
+
         user = self.connection.call("%s/%s" % (self.url, "who-am-i"))
-        self.log("Welcome %s" % (user["full_name"] or "user '%s'" % user["uid"],))
-        info = self.connection.call("%s/%s" % (self.url, "info")).get("info")
-        self.show_info(info, user["uid"])
-        info = self.connection.call("%s/%s/%s" % (self.url, self.service, "info")).get(
-            "info"
-        )
-        self.show_info(info, user["uid"])
+
+        if os.getenv("GITHUB_ACTION") is None:
+            self.log("Welcome %s" % (user["full_name"] or "user '%s'" % user["uid"],))
+
+        general_info = self.connection.call("%s/%s" % (self.url, "info")).get("info")
+        self.show_info(general_info, user["uid"])
+
+        service_specific_info = self.connection.call(
+            "%s/%s/%s" % (self.url, self.service, "info")
+        ).get("info")
+        self.show_info(service_specific_info, user["uid"])
 
         if news:
             try:
@@ -430,17 +467,6 @@ class APIRequest(object):
             except:
                 pass
 
-    def log(self, m):
-        if m == "":
-            if self._empty_line:
-                return
-            self._log(m)
-            self._empty_line = True
-            return
-
-        self._empty_line = False
-        self._log(m)
-
     def _bytename(self, size):
         prefix = {"": "K", "K": "M", "M": "G", "G": "T", "T": "P", "P": "E"}
         l = ""
@@ -490,7 +516,6 @@ class APIRequest(object):
         return existing_size + bytes_transferred
 
     def execute(self, request, target=None):
-
         status = None
 
         self.connection.submit("%s/%s/requests" % (self.url, self.service), request)
@@ -555,11 +580,10 @@ class APIRequest(object):
             self.log("")
 
 
-###############################################################################
-
-
 class ECMWFDataServer(object):
-    def __init__(self, url=None, key=None, email=None, verbose=False, log=None):
+    def __init__(
+        self, url=None, key=None, email=None, verbose=False, log=print_with_timestamp
+    ):
         if url is None or key is None or email is None:
             key, url, email = get_apikey_values()
 
@@ -569,30 +593,20 @@ class ECMWFDataServer(object):
         self.verbose = verbose
         self.log = log
 
-    def trace(self, m):
-        if self.log:
-            self.log(m)
-        else:
-            t = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
-            print("%s %s" % (t, m,))
-
     def retrieve(self, req):
         target = req.get("target")
         dataset = req.get("dataset")
         c = APIRequest(
             self.url,
             "datasets/%s" % (dataset,),
-            self.email,
-            self.key,
-            self.trace,
+            email=self.email,
+            key=self.key,
+            log=self.log,
             verbose=self.verbose,
         )
         c.execute(req, target)
 
 
-###############################################################################
-
-
 class ECMWFService(object):
     def __init__(
         self,
@@ -601,7 +615,7 @@ class ECMWFService(object):
         key=None,
         email=None,
         verbose=False,
-        log=None,
+        log=print_with_timestamp,
         quiet=False,
     ):
         if url is None or key is None or email is None:
@@ -615,24 +629,15 @@ class ECMWFService(object):
         self.quiet = quiet
         self.log = log
 
-    def trace(self, m):
-        if self.log:
-            self.log(m)
-        else:
-            t = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
-            print("%s %s" % (t, m,))
-
     def execute(self, req, target):
         c = APIRequest(
             self.url,
             "services/%s" % (self.service,),
-            self.email,
-            self.key,
-            self.trace,
+            email=self.email,
+            key=self.key,
+            log=self.log,
             verbose=self.verbose,
             quiet=self.quiet,
         )
         c.execute(req, target)
-        self.trace("Done.")
-
-    ###############################################################################
+        self.log("Done")
